#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Simple YAML emitter.
'''

# Python 2, 3 compatibility for print().
from __future__ import print_function


class YamlEmitter(object):
    '''
    Simple YAML emitter.
    '''
    def __init__(self):
        '''Initialization.
        '''
        self.envs = []
        self.indent = 0
        self.next_indent = self.indent
        self.key_is_next = True
        return

    def begin_doc(self):
        print('---')
        return

    def add_comment(self, comment):
        print(self.indent*' ' + '# ' + comment)
        return

    def begin_seq(self):
        if not self.envs:
            pass
        elif self.envs[-1] == 'seq':
            self.indent += 2
            self.next_indent = 0
            print('-', end=' ')
        elif self.envs[-1] == 'map':
            self.indent += 4
            self.next_indent = self.indent
            # print()
            self.key_is_next = True
        else:
            raise ValueError('Unknown environment.')
        self.envs.append('seq')
        return

    def add_item(self, item):
        assert self.envs
        assert self.envs[-1] == 'seq'
        print(self.next_indent*' ' + '-' + item)
        self.next_indent = self.indent
        return

    def end_seq(self):
        assert self.envs
        assert self.envs[-1] == 'seq'
        self.envs.pop()
        if not self.envs:
            pass
        elif self.envs[-1] == 'seq':
            self.indent -= 2
        elif self.envs[-1] == 'map':
            self.indent -= 4
        else:
            raise ValueError('Unknown environment.')
        self.next_indent = self.indent
        return

    def begin_map(self):
        if not self.envs:
            pass
        elif self.envs[-1] == 'seq':
            print(self.indent*' ' + '-', end=' ')
            self.indent += 2
            self.next_indent = 0
        elif self.envs[-1] == 'map':
            self.indent += 4
            self.next_indent = self.indent
            print()
        else:
            raise ValueError('Unknown environment.')
        self.envs.append('map')
        self.key_is_next = True
        return

    def add_key(self, key):
        assert self.envs
        assert self.envs[-1] == 'map'
        assert self.key_is_next
        print(self.next_indent*' ' + '%r:' % key)
        self.key_is_next = False
        return

    def add_value(self, item):
        assert self.envs
        assert self.envs[-1] == 'map'
        assert not self.key_is_next
        print('%r' % item)
        self.key_is_next = True
        self.next_indent = self.indent
        return

    def add_key_value(self, key, value):
        assert self.envs
        assert self.envs[-1] == 'map'
        assert self.key_is_next
        print(self.next_indent*' ' + '%r: %r' % (key, value))
        self.next_indent = self.indent
        return

    def end_map(self):
        assert self.envs
        assert self.envs[-1] == 'map'
        self.envs.pop()
        if not self.envs:
            pass
        elif self.envs[-1] == 'seq':
            self.indent -= 2
        elif self.envs[-1] == 'map':
            self.indent -= 4
        else:
            raise ValueError('Unknown environment.')
        self.next_indent = self.indent
        return



# -*- coding: utf-8 -*-
#
#  Copyright (c) 2012--2014, Nico Schlömer, <nico.schloemer@gmail.com>
#  All rights reserved.
#
#  This file is part of PyNosh.
#
#  PyNosh is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  PyNosh is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with PyNosh.  If not, see <http://www.gnu.org/licenses/>.
#
from . import modelevaluator_nls
from . import modelevaluator_bordering_constant
from . import numerical_methods
from . import preconditioners
from . import magnetic_vector_potentials
from . import yaml

__all__ = [
    'modelevaluator_nls',
    'modelevaluator_bordering_constant',
    'numerical_methods',
    'preconditioners',
    'magnetic_vector_potentials',
    'yaml'
    ]

__name__ = 'pynosh'
__version__ = '0.2.1'
__author__ = 'Nico Schlömer'
__author_email__ = 'nico.schloemer@gmail.com'



'''Module that provides magnetic vector potentials.'''
import numpy


def constant_field(X, B):
    '''Converts a spatially constant magnetic field B at X
    into a corresponding potential.
    '''
    # This is one particular choice that works.
    return 0.5 * numpy.cross(B, X)


def magnetic_dipole(x, x0, m):
    '''Magnetic vector potential for the static dipole at x0 with orientation
    m.
    '''
    r = x - x0
    # npsum(...) = ||r||^3 row-wise;
    # numpy.cross acts on rows by default;
    # The ".T" magic makes sure that each row of numpy.cross(m, r)
    # gets divided by the corresponding entry in ||r||^3.
    return (numpy.cross(m, r).T /
            numpy.sum(numpy.abs(r)**2, axis=-1)**(3./2)
            ).T


def magnetic_dot(X, radius, heights):
    '''Magnetic vector potential corresponding to the field that is induced by
    a cylindrical magnetic dot, centered at (0,0,0.5*(height0+height1)), with
    the radius `radius` for objects in the x-y-plane.  The potential is derived
    by interpreting the dot as an infinitesimal collection of magnetic dipoles,
    hence

       A(x) = \int_{dot} A_{dipole}(x-r) dr.

    Support for input valued (x,y,z), z!=0, is pending.
    '''
    # Span a cartesian grid over the sample, and integrate over it.
    # For symmetry, choose a number that is divided by 4.
    n_phi = 100
    # Choose such that the quads at radius/2 are approximately squares.
    n_radius = int(round(n_phi / numpy.pi))

    dr = radius / n_radius

    A = numpy.zeros((len(X), 3))

    # What we want to have is the value of
    #
    #    I(X) := \int_{dot} \|X-XX\|^{-3/2} (m\times(X-XX)) dXX
    #
    # with
    #
    #    X := (x, y, z)^T,
    #    XX := (xx, yy, zz)^T
    #
    # The integral in zz-direction (height) can be calculated analytically,
    # such that
    #
    #    I = \int_{disk}
    #          [ - (z-zz) / (r2D*sqrt(r3D)) ]_{zz=h_0}^{h_1}
    #          ( -(y-yy), x-xx, 0)^T dxx dyy.
    #
    # The integral over the disk is then approximated numerically by
    # the summation over little disk segments.
    # An alternative is to use cylindrical coordinates.
    #
    X_dist = numpy.empty((X.shape[0], 2))
    for i_phi in range(n_phi):
        beta = 2.0 * numpy.pi / n_phi * i_phi
        sin_beta = numpy.sin(beta)
        cos_beta = numpy.cos(beta)
        for i_radius in range(n_radius):
            rad = radius / n_radius * (i_radius + 0.5)
            # r = squared distance between grid point X to the
            #     point (x,y) on the magnetic dot
            X_dist[:, 0] = X[:, 0] - rad * cos_beta
            X_dist[:, 1] = X[:, 1] - rad * sin_beta

            # r = x_dist * x_dist + y_dist * y_dist
            # Note that X_dist indeed only has two components.
            R = numpy.sum(X_dist**2, axis=1)
            ind = numpy.nonzero(R > 1.0e-15)

            # 3D distance to point on lower edge (xi,yi,height0)
            # and upper edge ( (xi,yi,height1), respectively
            R_3D = [numpy.sqrt(R[ind] + heights[0]**2),
                    numpy.sqrt(R[ind] + heights[1]**2)
                    ]
            # Volume of circle segment = pi*angular_width * r^2,
            # so the volume of a building brick of the discretization is
            #   V = pi/n_phi * [(r+dr/2)^2 - (r-dr/2)^2]
            #     = pi/n_phi * 2 * r * dr.
            Alpha = (heights[1]/R_3D[1] - heights[0]/R_3D[0]) / R[ind] \
                * numpy.pi / n_phi * (2.0*rad*dr)  # volume
            # ax += y_dist * alpha
            # ay -= x_dist * alpha
            A[ind, 0] += X_dist[ind, 1] * Alpha
            A[ind, 1] -= X_dist[ind, 0] * Alpha
    return A



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Collection of numerical algorithms.
'''
import numpy
import krypy


class ForcingConstant(object):
    def __init__(self, eta0):
        self.eta0 = eta0
        return

    def get(self, eta_previous, resval_previous, F0, F_1):
        return self.eta0


class Forcing_EW1(object):
    '''Linear tolerance is given by

    "Choosing the Forcing Terms in an Inexact Newton Method (1994)"
    -- Eisenstat, Walker
    http://citeseer.ist.psu.edu/viewdoc/summary?doi=10.1.1.15.3196

    See also
    "NITSOL: A Newton Iterative Solver for Nonlinear Systems"
    http://epubs.siam.org/sisc/resource/1/sjoce3/v19/i1/p302_s1?isAuthorized=no
    '''
    def __init__(self, eta_min=1.0e-6, eta_max=1.0e-2):
        self.eta_min = eta_min
        self.eta_max = eta_max
        return

    def get(self, eta_previous, resval_previous, F0, F_1):
        from scipy.constants import golden
        # linear_relresvec[-1] \approx tol, so this could be replaced.
        eta = abs(F0 - resval_previous) / F_1
        eta = max(eta, eta_previous**golden, self.eta_min)
        eta = min(eta, self.eta_max)
        return eta


class Forcing_EW2(object):
    def __init__(self, eta_min=1.0e-6, eta_max=1.0e-2, alpha=1.5, gamma=0.9):
        self.eta_min = eta_min
        self.eta_max = eta_max
        self.alpha = alpha
        self.gamma = gamma
        return

    def get(self, eta_previous, resval_previous, F0, F_1):
        eta = self.gamma * (F0 / F_1)**self.alpha
        eta = max(eta, self.gamma * eta_previous**self.alpha, self.eta_min)
        eta = min(eta, self.eta_max)
        return eta


def newton(x0,
           model_evaluator,
           nonlinear_tol=1.0e-10,
           newton_maxiter=20,
           RecyclingSolver=krypy.recycling.RecyclingMinres,
           recycling_solver_kwargs=None,
           vector_factory_generator=None,
           compute_f_extra_args={},
           eta0=1.0e-10,
           forcing_term='constant',
           debug=False,
           yaml_emitter=None
           ):
    '''Newton's method with different forcing terms.
    '''

    # Default forcing term.
    if forcing_term == 'constant':
        forcing_term = ForcingConstant(eta0)

    if recycling_solver_kwargs is None:
        recycling_solver_kwargs = {}

    # Some initializations.
    # Set the default error code to 'failure'.
    error_code = 1
    k = 0

    x = x0.copy()
    Fx = model_evaluator.compute_f(x, **compute_f_extra_args)
    Fx_norms = [numpy.sqrt(model_evaluator.inner_product(Fx, Fx))]
    eta_previous = None
    linear_relresvecs = []

    # get recycling solver
    recycling_solver = RecyclingSolver()

    # no solution in before first iteration if Newton
    out = None

    if debug:
        import yaml
        if yaml_emitter is None:
            yaml_emitter = yaml.YamlEmitter()
            yaml_emitter.begin_doc()
        yaml_emitter.begin_seq()

    while Fx_norms[-1] > nonlinear_tol and k < newton_maxiter:
        if debug:
            yaml_emitter.add_comment('Newton step %d' % (k+1))
            yaml_emitter.begin_map()
            yaml_emitter.add_key_value('Fx_norm', Fx_norms[-1][0][0])

        # Get tolerance for next linear solve.
        if k == 0:
            eta = eta0
        else:
            eta = forcing_term.get(eta_previous, out.resnorms[-1],
                                   Fx_norms[-1], Fx_norms[-2]
                                   )
        eta_previous = eta

        # Setup linear problem.
        jacobian = model_evaluator.get_jacobian(x, **compute_f_extra_args)

        M = model_evaluator.get_preconditioner(x, **compute_f_extra_args)
        Minv = \
            model_evaluator.get_preconditioner_inverse(x,
                                                       **compute_f_extra_args
                                                       )

        # get vector factory
        if vector_factory_generator is not None:
            vector_factory = vector_factory_generator(x)
        else:
            vector_factory = None

        # Create the linear system.
        linear_system = krypy.linsys.TimedLinearSystem(
            jacobian, -Fx, M=Minv, Minv=M, ip_B=model_evaluator.inner_product,
            normal=True, self_adjoint=True
            )

        out = recycling_solver.solve(linear_system,
                                     vector_factory,
                                     tol=eta,
                                     **recycling_solver_kwargs
                                     )

        if debug:
            yaml_emitter.add_key_value('relresvec', out.resnorms)
            #yaml_emitter.add_key_value('relresvec[-1]', out['relresvec'][-1])
            yaml_emitter.add_key_value('num_iter', len(out.resnorms)-1)
            yaml_emitter.add_key_value('eta', eta)

        # save the convergence history
        linear_relresvecs.append(out.resnorms)

        # perform the Newton update
        x += out.xk

        # do the household
        k += 1
        Fx = model_evaluator.compute_f(x, **compute_f_extra_args)
        Fx_norms.append(numpy.sqrt(model_evaluator.inner_product(Fx, Fx)))

        # run garbage collector in order to prevent MemoryErrors from being
        # raised
        import gc
        gc.collect()

        if debug:
            yaml_emitter.end_map()

    if Fx_norms[-1] < nonlinear_tol:
        error_code = 0

    if debug:
        yaml_emitter.begin_map()
        yaml_emitter.add_key_value('Fx_norm', Fx_norms[-1])
        yaml_emitter.end_map()
        yaml_emitter.end_seq()
        if Fx_norms[-1] > nonlinear_tol:
            yaml_emitter.add_comment(
                'Newton solver did not converge '
                '(residual = %g > %g = tol)' % (Fx_norms[-1], nonlinear_tol)
                )

    return {'x': x,
            'info': error_code,
            'Newton residuals': Fx_norms,
            'linear relresvecs': linear_relresvecs,
            'recycling_solver': recycling_solver
            }


def poor_mans_continuation(x0,
                           model_evaluator,
                           initial_parameter_value,
                           initial_step_size=1.0e-2,
                           minimal_step_size=1.0e-6,
                           maximum_step_size=1.0e-1,
                           max_steps=1000,
                           nonlinear_tol=1.0e-10,
                           max_newton_iters=5,
                           adaptivity_aggressiveness=1.0
                           ):
    '''Poor man's parameter continuation. With adaptive step size.
    If the previous step was unsuccessful, the step size is cut in half,
    but if the step was successful this strategy increases the step size based
    on the number of nonlinear solver iterations required in the previous step.
    In particular, the new step size :math:`\Delta s_{new}` is given by

    .. math::
       \Delta s_{new} = \Delta s_{old}\left(1 + a\left(
         \\frac{N_{max} - N}{N_{max}}\\right)^2\\right).
    '''

    # write header of the statistics file
    stats_file = open('continuationData.dat', 'w')
    stats_file.write('# step    parameter     norm            Newton iters\n')
    stats_file.flush()

    parameter_value = initial_parameter_value
    x = x0

    current_step_size = initial_step_size

    for k in range(max_steps):
        print('Continuation step %d (parameter=%e)...' % (k, parameter_value))

        # Try to converge to a solution and adapt the step size.
        converged = False
        while current_step_size > minimal_step_size:
            x_new, error_code, iters = newton(x,
                                              model_evaluator,
                                              nonlinear_tol=nonlinear_tol,
                                              max_iters=max_newton_iters
                                              )
            if error_code != 0:
                current_step_size *= 0.5
                print(('Continuation step failed (error code %d). '
                       'Setting step size to %e.')
                      % (error_code, current_step_size)
                      )
            else:
                current_step_size *= 1.0 + adaptivity_aggressiveness * \
                    (float(max_newton_iters-iters)/max_newton_iters)**2
                converged = True
                x = x_new
                print('Continuation step success!')
                break

        if not converged:
            print('Could not find a solution although '
                  'the step size was %e. Abort.' % current_step_size
                  )
            break

        stats_file.write('  %4d    %.5e   %.5e    %d\n' %
                         (k, parameter_value, model_evaluator.energy(x), iters)
                         )
        stats_file.flush()
        #model_evaluator.write( x, "step" + str(k) + ".vtu" )

        parameter_value += current_step_size
        model_evaluator.set_parameter(parameter_value)

    stats_file.close()

    print('done.')
    return



# -*- coding: utf-8 -*-
import numpy
from scipy.sparse.linalg import LinearOperator
import krypy


class ConstBorderedModelEvaluator(object):
    '''
    Wraps a given model evaluator in a bordering strategy. Does not work with
    preconditioners.
    '''
    def __init__(self, modeleval, bord):
        '''Initialization.
        '''
        self.inner_modeleval = modeleval
        self.dtype = self.inner_modeleval.dtype
        self.bord = bord
        return

    def compute_f(self, x):
        '''Compute bordered F.
        '''
        n = len(x) - 1
        res = numpy.empty((n+1, 1), dtype=self.dtype)
        # Right border: bord * eta.
        res[0:n] = self.inner_modeleval.compute_f(x[0:n]) \
            + self.bord * x[n]
        # Lower border: <bord, psi>.
        res[n] = self.inner_modeleval.inner_product(self.bord, x[0:n])
        return res

    def get_jacobian(self, x0):
        '''Jacobian of the bordered system.
        '''
        def _apply_jacobian(x):
            y = numpy.empty(x.shape, dtype=self.dtype)
            y[0:n] = inner_jacobian * x[0:n] + b.reshape(x[0:n].shape) * x[n]
            assert abs(x[n].imag) < 1.0e-15, 'Not real-valued: %r.' % x[n]
            y[n] = self.inner_modeleval.inner_product(c, x[0:n]) \
                + d * x[n].real
            return y

        assert x0 is not None

        n = len(x0) - 1
        psi0 = x0[0:n]

        b = self.bord
        c = self.bord
        d = 0.0

        inner_jacobian = self.inner_modeleval.get_jacobian(psi0)
        return LinearOperator((n+1, n+1),
                              _apply_jacobian,
                              dtype=self.dtype
                              )

    def get_jacobian_inverse(self, x0):
        '''Preconditioner based on Schur complement.
        '''
        def _apply_jacobian_inverse(x):
            '''Schur thing.'''
            # Test: Use with Jacobian solved exactly.
            phi0 = x[0:n]
            phi0 = phi0.reshape((n, 1))

            linear_system0 = krypy.linsys.LinearSystem(
                jacobian, phi0,
                ip_B=self.inner_modeleval.inner_product
                )
            out0 = krypy.linalg.Minres(linear_system0,
                                       tol=1.0e-11,
                                       maxiter=500
                                       )
            z0 = out0.xk.reshape(n)

            linear_system1 = krypy.linsys.LinearSystem(
                jacobian, b,
                ip_B=self.inner_modeleval.inner_product
                )
            out1 = krypy.linalg.Minres(linear_system1,
                                       tol=1.0e-11,
                                       maxiter=500
                                       )
            z1 = out1.xk.reshape(n)

            # Schur complement.
            s = d - self.inner_modeleval.inner_product(c, z1)
            assert abs(s) > 1.0e-15
            tmp = self.inner_modeleval.inner_product(c, z0)

            assert abs(s.imag) < 1.0e-15
            s = s.real
            assert abs(tmp.imag) < 1.0e-15
            tmp = tmp.real
            assert abs(x[n].imag) < 1.0e-15
            xn = x[n].real

            y = numpy.empty(x.shape, dtype=self.dtype)
            y[0:n] = z0 + z1 * 1.0/s * tmp - z1 / s * xn
            y[n] = - 1.0/s * tmp + 1.0/s * xn
            return y

        n = len(x0) - 1
        psi0 = x0[0:n]

        b = self.bord
        c = self.bord
        d = 0.0

        jacobian = self.inner_modeleval.get_jacobian(psi0)
        #prec = self.inner_modeleval.get_preconditioner_inverse(psi0)
        return LinearOperator((n+1, n+1),
                              _apply_jacobian_inverse,
                              dtype=self.dtype
                              )

    def get_preconditioner(self, psi0):
        # Preconditioner not supported.
        return None

    def get_preconditioner_inverse(self, x0):
        '''Preconditioner based on Schur complement.
        '''
        def _apply_precon_inverse(x):
            '''Schur thing.'''
            # Approximate solutions of
            #    J^{-1} phi
            # and
            #    J^{-1} b.
            z0 = prec * x[0:n]
            z1 = prec * b.reshape(z0.shape)

            # Schur complement.
            s = d - self.inner_modeleval.inner_product(c, z1)
            assert abs(s) > 1.0e-15
            tmp = self.inner_modeleval.inner_product(c, z0)

            assert abs(s.imag) < 1.0e-15
            s = s.real
            assert abs(tmp.imag) < 1.0e-15
            tmp = tmp.real
            assert abs(x[n].imag) < 1.0e-15
            xn = x[n].real

            y = numpy.empty(x.shape, dtype=self.dtype)
            y[0:n] = z0 + z1 * 1.0/s * tmp - z1 / s * xn
            y[n] = - 1.0/s * tmp + 1.0/s * xn
            return y

        n = len(x0) - 1
        psi0 = x0[0:n]

        b = self.bord
        c = self.bord
        d = 0.0

        prec = self.inner_modeleval.get_preconditioner_inverse(psi0)
        if prec:
            return LinearOperator((n+1, n+1),
                                  _apply_precon_inverse,
                                  dtype=self.dtype
                                  )
        else:
            return None

    def inner_product(self, x0, x1):
        '''The inner product of the bordered problem.
        '''
        n = len(x0) - 1
        return self.inner_modeleval.inner_product(x0[0:n], x1[0:n]) \
            + x0[n] * x1[n]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Preconditioners for the Jacobian of the Ginzburg--Landau problem.
'''
from scipy import sparse
from scipy.sparse.linalg import splu, spilu
import pynosh.numerical_methods as nm


class Preconditioners(object):
    '''
    Ginzburg--Landau model evaluator class.
    '''
    def __init__(self, model_evaluator):
        '''Initialization.
        '''
        self._modeleval = model_evaluator
        self._keo_lu = None
        self._keo_ilu = None
        self._keo_ilu_droptol = None
        self._keo_symmetric_ilu = None
        self._keo_symmetric_ilu_droptol = None
        self._keoai_lu = None
        self._keo_amg_solver = None
        return

    def diagonal(self, psi):
        '''The equivalent of a diagonal preconditioner.
        Solves the equation system with

            self._keo.diagonal() * x \
            - ( 1.0 - 2.0*abs(current_psi)**2 ) * x \
            + current_psi**2 * x.conjugate()
            = psi.
        '''
        if self._modeleval._keo is None:
            self._modeleval._assemble_kinetic_energy_operator()

        # Mind that all operations below are executed elementwise.
        a = self._modeleval._keo.diagonal() \
            - (1.0-self._modeleval._temperature -
               2.0*abs(self._modeleval._psi)**2
               )
        b = self._modeleval._psi**2

        # One needs to solve
        #    a*z + b z.conj = psi
        # for each "diagonal" entry z. The solution is
        #   z = ( a.conj * psi - b * psi.conj ) / ( |a|^2 - |b|^2 )
        alpha = abs(a)**2 - abs(b)**2
        assert((abs(alpha) > 1.0e-10).all)
        return (a.conjugate() * psi - b * psi.conjugate()) / alpha

    def keo_cgapprox(self, psi):
        '''Solves a system with the kinetic energy operator only using
        ordinary CG.
        '''
        if self._modeleval._keo is None:
            self._modeleval._assemble_kinetic_energy_operator()

        sol, info = nm.cg_wrap(self._modeleval._keo,
                               psi,
                               x0=None,
                               inner_product=self._modeleval.inner_product,
                               tol=1.0e-10,
                               M=None,
                               callback=None
                               )
        # make sure it could be solved
        assert info == 0
        return sol

    def keo_amg(self, psi):
        '''Algebraic multigrid solve.
        '''
        import pyamg
        if self._keo_amg_solver is None:
            if self._modeleval._keo is None:
                self._modeleval._assemble_keo()
            self._keo_amg_solver = \
                pyamg.smoothed_aggregation_solver(self._modeleval._keo)
        return self._keo_amg_solver.solve(psi,
                                          tol=1e-12,
                                          accel=None
                                          )

    def keo_lu(self, psi):
        '''Solves a system with the kinetic energy operator only via ordinary
        CG.
        '''
        if self._modeleval._keo is None:
            self._modeleval._assemble_kinetic_energy_operator()

        # From http://crd.lbl.gov/~xiaoye/SuperLU/faq.html#sym-problem:
        # SuperLU cannot take advantage of symmetry, but it can still solve the
        # linear system as long as you input both the lower and upper parts of
        # the matrix A. If off-diagonal pivoting does not occur, the U matrix
        # in A = L*U is equivalent to D*L'.
        # In many applications, matrix A may be diagonally dominant or nearly
        # so. In this case, pivoting on the diagonal is sufficient for
        # stability and is preferable for sparsity to off-diagonal pivoting. To
        # do this, the user can set a small (less-than-one) diagonal pivot
        # threshold (e.g., 0.0, 0.01, ...) and choose an (A' + A)-based column
        # permutation algorithm. We call this setting Symmetric  Mode. To use
        # this (in serial SuperLU), you need to set:
        #
        #    options.SymmetricMode = YES;
        #    options.ColPerm = MMD_AT_PLUS_A;
        #    options.DiagPivotThresh = 0.001; /* or 0.0, 0.01, etc. */
        #
        if self._keo_lu is None:
            self._keo_lu = splu(self._modeleval._keo,
                                options={'SymmetricMode': True},
                                permc_spec='MMD_AT_PLUS_A',  # minimum deg
                                diag_pivot_thresh=0.0
                                )
        return self._keo_lu.solve(psi)

    def keo_symmetric_ilu2(self, psi):
        return self.keo_symmetric_ilu(psi, 1.0e-2)

    def keo_symmetric_ilu4(self, psi):
        return self.keo_symmetric_ilu(psi, 1.0e-4)

    def keo_symmetric_ilu6(self, psi):
        return self.keo_symmetric_ilu(psi, 1.0e-6)

    def keo_symmetric_ilu8(self, psi):
        return self.keo_symmetric_ilu(psi, 1.0e-8)

    def keo_symmetric_ilu(self, psi, droptol):
        if self._modeleval._keo is None:
            self._modeleval._assemble_kinetic_energy_operator()
        if self._keo_symmetric_ilu is None \
           or self._keo_symmetric_ilu_droptol is None \
           or droptol != self._keo_symmetric_ilu_droptol:
            self._keo_symmetric_ilu = spilu(self._modeleval._keo,
                                            drop_tol=droptol,
                                            fill_factor=10,
                                            drop_rule=None,
                                            # see remark above for splu
                                            options={'SymmetricMode': True},
                                            permc_spec='MMD_AT_PLUS_A',
                                            diag_pivot_thresh=0.0,
                                            relax=None,
                                            panel_size=None
                                            )
            self._keo_symmetric_ilu_droptol = droptol
        return self._keo_symmetric_ilu.solve(psi)

    def keo_ilu4(self, psi):
        return self.keo_ilu(psi, 1.0e-4)

    def keo_ilu6(self, psi):
        return self.keo_ilu(psi, 1.0e-6)

    def keo_ilu(self, psi, droptol):
        if self._modeleval._keo is None:
            self._modeleval._assemble_kinetic_energy_operator()
        if self._keo_ilu is None \
           or self._keo_ilu_droptol is None \
           or droptol != self._keo_ilu_droptol:
            self._keo_ilu = spilu(self._modeleval._keo,
                                  drop_tol=droptol,
                                  fill_factor=10,
                                  drop_rule=None,
                                  relax=None,
                                  panel_size=None
                                  )
            self._keo_ilu_droptol = droptol
        return self._keo_ilu.solve(psi)

    def keoi(self, psi):
        if self._modeleval._keo is None:
            self._modeleval._assemble_kinetic_energy_operator()
        if self._keoai_lu is None:
            self._keoai_lu = splu(
                self._modeleval._keo + 1.0e-2 * sparse.identity(len(psi))
                )
        return self._keoai_lu.solve(psi)

    def set_parameter(self, mu):
        '''This function is actually somewhat funny in that it is not used to
        actually set mu (which lives in the model evaluator anyway), but to
        make sure that all the corresponding objects (the factorizations) are
        recreated next time they are needed.
        '''
        self._modeleval.set_parameter(mu)
        self._keo_lu = None
        self._keo_ilu = None
        self._keo_ilu_droptol = None
        self._keo_symmetric_ilu = None
        self._keo_symmetric_ilu_droptol = None
        self._keoai_lu = None
        self._keo_amg_solver = None
        return

    def set_mesh(self, mesh):
        '''This function is actually somewhat funny in that it is not used to
        actually set the mesh (which lives in the model evaluator anyway), but
        to make sure that all the corresponding objects (the factorizations)
        are recreated next time they are needed.
        '''
        self._modeleval.set_mesh(mesh)
        self._keo_lu = None
        self._keo_ilu = None
        self._keo_ilu_droptol = None
        self._keo_symmetric_ilu = None
        self._keo_symmetric_ilu_droptol = None
        self._keoai_lu = None
        self._keo_amg_solver = None
        return



# -*- coding: utf-8 -*-
#
#  Copyright (c) 2012--2014, Nico Schlömer, <nico.schloemer@gmail.com>
#  All rights reserved.
#
#  This file is part of PyNosh.
#
#  PyNosh is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  PyNosh is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with PyNosh.  If not, see <http://www.gnu.org/licenses/>.
#
import os
from distutils.core import setup
import codecs

from pynosh import __version__, __author__, __author_email__


def read(fname):
    try:
        content = codecs.open(
            os.path.join(os.path.dirname(__file__), fname),
            encoding='utf-8'
            ).read()
    except Exception:
        content = ''
    return content


setup(
    name='pynosh',
    packages=['pynosh'],
    version=__version__,
    description='Nonlinear Schrödinger equations',
    long_description=read('README.rst'),
    author=__author__,
    author_email=__author_email__,
    url='https://github.com/nschloe/pynosh/',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Mathematics'
        ],
    )



# -*- coding: utf-8 -*-
#
#  Copyright (c) 2012--2014, Nico Schlömer, <nico.schloemer@gmail.com>
#  All rights reserved.
#
#  This file is part of PyNosh.
#
#  PyNosh is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  PyNosh is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with PyNosh.  If not, see <http://www.gnu.org/licenses/>.
#
'''
Provide information around the nonlinear Schrödinger equations.
'''
import numpy
from scipy import sparse, linalg
import warnings
import krypy


class NlsModelEvaluator(object):
    '''Nonlinear Schrödinger model evaluator class.
    Incorporates

       * Nonlinear Schrödinger: :math:`g=1.0, V=0.0, A=0.0`.
       * Gross--Pitaevskii: :math:`g=1.0`, :math:`V` given, :math:`A=0.0`.
       * Ginzburg--Landau: :math:`g=1.0, V=-1.0`,
         and some magnetic potential :math:`A`.
    '''
    def __init__(self,
                 mesh,
                 V=None,
                 A=None,
                 preconditioner_type='none',
                 num_amg_cycles=numpy.inf
                 ):
        '''Initialization. Set mesh.
        '''
        self.dtype = complex
        self.mesh = mesh
        n = len(mesh.node_coords)
        if V is None:
            self._V = numpy.zeros(n)
        else:
            self._V = V
        if A is None:
            self._raw_magnetic_vector_potential = numpy.zeros((n, 3))
        else:
            self._raw_magnetic_vector_potential = A
        self._keo_cache = None
        self._keo_cache_mu = 0.0
        self._edgecoeff_cache = None
        self.tot_amg_cycles = []
        self.cv_variant = 'voronoi'
        self._preconditioner_type = preconditioner_type
        self._num_amg_cycles = num_amg_cycles
        return

    def compute_f(self, x, mu, g):
        '''Computes the nonlinear Schrödinger residual

        .. math::
            GP(\\psi) = K\\psi + (V + g |\\psi|^2) \\psi
        '''
        keo = self._get_keo(mu)
        if self.mesh.control_volumes is None:
            self.mesh.compute_control_volumes(variant=self.cv_variant)
        res = (keo * x) / self.mesh.control_volumes.reshape(x.shape) \
            + (self._V.reshape(x.shape) + g * abs(x)**2) * x
        return res

    def get_jacobian(self, x, mu, g):
        '''Returns a LinearOperator object that defines the matrix-vector
        multiplication scheme for the Jacobian operator as in

        .. math::
            A \\phi + B \\phi^*

        with

        .. math::
            A &= K + I (V + g \\cdot 2|\\psi|^2),\\\\
            B &= g \\cdot  diag( \\psi^2 ).
        '''
        def _apply_jacobian(phi):
            if len(phi.shape) == 1:
                shape = phi.shape
            elif len(phi.shape) == 2:
                # phi may be a vector of shape (n, k).
                shape = (phi.shape[0], 1)
            else:
                raise ValueError('Illegal phi.')
            y = (keo * phi) / self.mesh.control_volumes.reshape(shape) \
                + alpha.reshape(shape) * phi \
                + gPsi0Squared.reshape(shape) * phi.conj()
            return y

        assert x is not None

        keo = self._get_keo(mu)

        if self.mesh.control_volumes is None:
            self.mesh.compute_control_volumes(variant=self.cv_variant)
        alpha = self._V.reshape(x.shape) + g * 2.0*(x.real**2 + x.imag**2)
        gPsi0Squared = g * x**2

        num_unknowns = len(self.mesh.node_coords)

        return krypy.utils.LinearOperator((num_unknowns, num_unknowns),
                                          self.dtype,
                                          dot=_apply_jacobian,
                                          dot_adj=_apply_jacobian
                                          )

    def get_jacobian_blocks(self, x, mu, g):
        '''Returns

        .. math::
            A &= K + I  (V + g \\cdot 2|\\psi|^2),\\\\
            B &= g \\cdot diag( \\psi^2 ).
        '''
        assert x is not None

        if self.mesh.control_volumes is None:
            self.mesh.compute_control_volumes(variant=self.cv_variant)

        A = self._get_keo(mu).copy()
        diag = A.diagonal()
        alpha = self._V.reshape(x.shape) + g * 2.0 * (x.real**2 + x.imag**2)
        diag += alpha.reshape(diag.shape) \
            * self.mesh.control_volumes.reshape(x.shape)
        A.setdiag(diag)

        num_nodes = len(self.mesh.node_coords)
        from scipy.sparse import spdiags
        B = spdiags(g * x**2 * self.mesh.control_volumes.reshape(x.shape),
                    [0],
                    num_nodes, num_nodes)
        return A, B

    def get_preconditioner(self, x, mu, g):
        '''Return the preconditioner.
        '''
        if self._preconditioner_type == 'none':
            return None
        if self._preconditioner_type == 'cycles':
            warnings.warn(
                'Preconditioner inverted approximately with '
                '%d AMG cycles, so get_preconditioner() isn\'t exact.'
                % self._num_amg_cycles
                )

        def _apply_precon(phi):
            return (keo * phi) / self.mesh.control_volumes.reshape(phi.shape) \
                + alpha.reshape(phi.shape) * phi
                # + beta.reshape(phi.shape) * phi.conj()

        assert x is not None

        keo = self._get_keo(mu)

        if self.mesh.control_volumes is None:
            self.mesh.compute_control_volumes(variant=self.cv_variant)

        if g > 0.0:
            alpha = g * 2.0 * (x.real**2 + x.imag**2)
            # beta = g * x**2
        else:
            alpha = numpy.zeros(len(x))
        num_unknowns = len(self.mesh.node_coords)
        return krypy.utils.LinearOperator((num_unknowns, num_unknowns),
                                          self.dtype,
                                          dot=_apply_precon
                                          )

    def get_preconditioner_inverse(self, x, mu, g):
        '''Use AMG to invert M approximately.
        '''
        if self._preconditioner_type == 'none':
            return None
        import pyamg

        num_unknowns = len(x)

        def _apply_inverse_prec_exact(phi):
            assert(len(phi.shape) == 2)
            assert(len(self.mesh.control_volumes.shape) == 1)
            rhs = numpy.empty(phi.shape, dtype=phi.dtype)
            sol = numpy.empty(phi.shape, dtype=phi.dtype)
            for i in range(phi.shape[1]):
                rhs = self.mesh.control_volumes * phi[:, i]
                linear_system = krypy.linsys.LinearSystem(
                    prec,
                    rhs,
                    M=amg_prec,
                    self_adjoint=True,
                    positive_definite=True
                    )
                x_init = numpy.zeros((num_unknowns, 1), dtype=complex)
                out = krypy.linsys.Cg(linear_system,
                                      x0=x_init,
                                      tol=1.0e-13,
                                      # explicit_residual = False
                                      )
                sol[:, i] = out.xk[:, 0]
                # Forget about the cycle used to gauge the residual norm.
                self.tot_amg_cycles += [len(out.resnorms) - 1]
            return sol

        def _apply_inverse_prec_cycles(phi):
            rhs = self.mesh.control_volumes.reshape((phi.shape[0], 1)) * phi
            x_init = numpy.zeros((num_unknowns, 1), dtype=complex)
            x = numpy.empty(phi.shape, dtype=complex)
            residuals = []
            for i in range(rhs.shape[1]):
                x[:, i] = prec_amg_solver.solve(rhs[:, i],
                                                x0=x_init,
                                                maxiter=self._num_amg_cycles,
                                                tol=0.0,
                                                accel=None,
                                                residuals=residuals
                                                )
            # Alternative for one cycle:
            # amg_prec = prec_amg_solver.aspreconditioner( cycle='V' )
            # x = amg_prec * rhs
            self.tot_amg_cycles += [self._num_amg_cycles]
            return x

        keo = self._get_keo(mu)

        if g > 0.0:
            if self.mesh.control_volumes is None:
                self.mesh.compute_control_volumes(variant=self.cv_variant)
            # don't use .setdiag,
            # cf. https://github.com/scipy/scipy/issues/3501
            alpha = g * 2.0 * (x.real**2 + x.imag**2) \
                * self.mesh.control_volumes.reshape(x.shape)
            prec = keo \
                + sparse.spdiags(alpha[:, 0], [0], num_unknowns, num_unknowns)
        else:
            prec = keo

        # The preconditioner assumes the eigenvalue 0 iff mu=0 and psi=0.
        # This may lead to problems if mu=0 and the Newton iteration
        # converges to psi=0 for psi0 != 0.
        # import scipy.sparse.linalg
        # lambd, v = scipy.sparse.linalg.eigs(prec, which='SM')
        # assert all(abs(lambd.imag) < 1.0e-15)
        # print '||psi||^2 = %g' % numpy.linalg.norm(absPsi0Squared)
        # print 'lambda =', lambd.real

        prec_amg_solver = \
            pyamg.smoothed_aggregation_solver(
                prec,
                strength=('evolution',
                          {'epsilon': 4.0, 'k': 2, 'proj_type': 'l2'}
                          ),
                smooth=('energy',
                        {'weighting': 'local',
                         'krylov': 'cg',
                         'degree': 2,
                         'maxiter': 3
                         }),
                improve_candidates=None,
                aggregate='standard',
                presmoother=('block_gauss_seidel',
                             {'sweep': 'symmetric', 'iterations': 1}
                             ),
                postsmoother=('block_gauss_seidel',
                              {'sweep': 'symmetric', 'iterations': 1}
                              ),
                max_levels=25,
                coarse_solver='splu'
                )

        # print 'operator complexity', prec_amg_solver.operator_complexity()
        # print 'cycle complexity', prec_amg_solver.cycle_complexity('V')

        if self._preconditioner_type == 'cycles':
            if self._num_amg_cycles == numpy.inf:
                raise ValueError('Invalid number of cycles.')
            return krypy.utils.LinearOperator((num_unknowns, num_unknowns),
                                              self.dtype,
                                              dot=_apply_inverse_prec_cycles
                                              )
        elif self._preconditioner_type == 'exact':
            amg_prec = prec_amg_solver.aspreconditioner(cycle='V')
            return krypy.utils.LinearOperator((num_unknowns, num_unknowns),
                                              dtype=self.dtype,
                                              dot=_apply_inverse_prec_exact
                                              )
        else:
            raise ValueError('Unknown preconditioner type ''%s''.'
                             % self._preconditioner_type
                             )

    def _get_preconditioner_inverse_directsolve(self, x, mu, g):
        '''Use a direct solver for M^{-1}.
        '''
        from scipy.sparse.linalg import spsolve

        def _apply_inverse_prec(phi):
            return spsolve(prec, phi)
        prec = self.get_preconditioner(x, mu, g)
        num_unknowns = len(x)
        return krypy.Utils.LinearOperator(
            (num_unknowns, num_unknowns),
            _apply_inverse_prec,
            dtype=self.dtype
            )

    def inner_product(self, phi0, phi1):
        '''The natural inner product of the problem.
        '''
        assert phi0.shape[0] == phi1.shape[0], \
            ('Input vectors not matching.', phi0.shape, phi1.shape)
        if self.mesh.control_volumes is None:
            self.mesh.compute_control_volumes(variant=self.cv_variant)
        if len(phi0.shape) == 1:
            scaledPhi0 = self.mesh.control_volumes * phi0
        elif len(phi0.shape) == 2:
            scaledPhi0 = \
                self.mesh.control_volumes.reshape((phi0.shape[0], 1)) * phi0
        # numpy.vdot only works for vectors, so use numpy.dot(....T.conj())
        # here.
        return numpy.dot(scaledPhi0.T.conj(), phi1).real

    def energy(self, psi):
        '''Compute the Gibbs free energy.
        Not really a norm, but a good measure for our purposes here.
        '''
        if self.mesh.control_volumes is None:
            self.mesh.compute_control_volumes(variant=self.cv_variant)
        alpha = -self.inner_product(psi**2, psi**2)
        return alpha.real / self.mesh.control_volumes.sum()

    def _get_keo(self, mu):
        '''Assemble the kinetic energy operator.'''

        if self._keo_cache is None or self._keo_cache_mu != mu:
            # Create the matrix structure.
            num_nodes = len(self.mesh.node_coords)

            mvp_edge_cache = self._build_mvp_edge_cache(mu)

            # Build caches.
            if self._edgecoeff_cache is None:
                self._build_edgecoeff_cache()
            if self.mesh.edges is None:
                self.mesh.create_adjacent_entities()

            n_edges = len(self.mesh.edges['nodes'])
            row = numpy.zeros(4*n_edges, dtype=int)
            col = numpy.zeros(4*n_edges, dtype=int)
            data = numpy.zeros(4*n_edges, dtype=complex)

            # loop over all edges
            for k, node_indices in enumerate(self.mesh.edges['nodes']):
                # Fetch the cached values.
                alpha = self._edgecoeff_cache[k]
                alphaExp0 = self._edgecoeff_cache[k] \
                    * numpy.exp(1j * mvp_edge_cache[k])
                # Sum them into the matrix.
                row[4*k:4*k+4] = [node_indices[0], node_indices[0],
                                  node_indices[1], node_indices[1]]
                col[4*k:4*k+4] = [node_indices[0], node_indices[1],
                                  node_indices[0], node_indices[1]]
                data[4*k:4*k+4] = [alpha, -alphaExp0.conj(),
                                   -alphaExp0, alpha]

            self._keo_cache = sparse.csr_matrix((data, (row, col)),
                                                (num_nodes, num_nodes))
            # transform the matrix into the more efficient CSR format
            self._keo_cache_mu = mu
        return self._keo_cache

    def _build_edgecoeff_cache(self):
        '''Build cache for the edge coefficients.
        (in 2D: coedge-edge ratios).
        '''
        # make sure the mesh has edges
        if self.mesh.edges is None:
            self.mesh.create_adjacent_entities()

        num_edges = len(self.mesh.edges)
        self._edgecoeff_cache = numpy.zeros(num_edges, dtype=float)

        if self.mesh.cell_volumes is None:
            self.mesh.create_cell_volumes()
        vols = self.mesh.cell_volumes

        # Precompute edges.
        edges = self.mesh.node_coords[self.mesh.edges['nodes'][:, 1]] \
            - self.mesh.node_coords[self.mesh.edges['nodes'][:, 0]]

        # Calculate the edge contributions cell by cell.
        for vol, cell in zip(vols, self.mesh.cells):
            cell_edge_gids = cell['edges']
            # Build the equation system:
            # The equation
            #
            # |simplex| ||u||^2 = \sum_i \alpha_i <u,e_i> <e_i,u>
            #
            # has to hold for all vectors u in the plane spanned by the edges,
            # particularly by the edges themselves.
            A = numpy.dot(edges[cell_edge_gids], edges[cell_edge_gids].T)
            # Careful here! As of NumPy 1.7, numpy.diag() returns a view.
            rhs = vol * numpy.diag(A).copy()
            A = A**2

            # Append the the resulting coefficients to the coefficient cache.
            # The system is posdef iff the simplex isn't degenerate.
            try:
                self._edgecoeff_cache[cell_edge_gids] += \
                    linalg.solve(A, rhs, sym_pos=True)
            except numpy.linalg.linalg.LinAlgError:
                # The matrix A appears to be singular,
                # and the only circumstance that makes this
                # happening is the cell being degenerate.
                # Hence, it has volume 0, and so all the edge
                # coefficients are 0, too.
                # Hence, do nothing.
                pass
        return

    def _build_mvp_edge_cache(self, mu):
        '''Builds the cache for the magnetic vector potential.'''

        # make sure the mesh has edges
        if self.mesh.edges is None:
            self.mesh.create_adjacent_entities()

        # Approximate the integral
        #
        #    I = \int_{x0}^{xj} (xj-x0)/|xj-x0| . A(x) dx
        #
        # numerically by the midpoint rule, i.e.,
        #
        #    I ~ (xj-x0) . A( 0.5*(xj+x0) )
        #      ~ (xj-x0) . 0.5*( A(xj) + A(x0) )
        #
        # The following computes the dot-products of all those
        # edges[i], mvp[i], and put the result in the cache.
        edges = self.mesh.node_coords[self.mesh.edges['nodes'][:, 1]] \
            - self.mesh.node_coords[self.mesh.edges['nodes'][:, 0]]
        mvp = 0.5 * (self._get_mvp(mu, self.mesh.edges['nodes'][:, 1]) +
                     self._get_mvp(mu, self.mesh.edges['nodes'][:, 0])
                     )
        return numpy.sum(edges * mvp, 1)

    def _get_mvp(self, mu, index):
        return mu * self._raw_magnetic_vector_potential[index]

    # def keo_smallest_eigenvalue_approximation(self):
    #     '''Returns
    #        <v,Av> / <v,v>
    #     with v = ones and A = KEO - Laplace.
    #     This is linear approximation for the smallest magnitude eigenvalue
    #     of KEO.
    #     '''
    #     num_nodes = len(self.mesh.nodes)

    #     # compute the FVM entities for the mesh
    #     if self._edge_lengths is None or self._coedge_edge_ratios is None:
    #         self._create_fvm_entities()

    #     k = 0
    #     sum = 0.0
    #     for element in self.mesh.cells:
    #         # loop over the edges
    #         l = 0
    #         for edge in element.edges:
    #             # -----------------------------------------------------------
    #             # Compute the integral
    #             #
    #             #    I = \int_{x0}^{xj} (xj-x0).A(x) dx
    #             #
    #             # numerically by the midpoint rule, i.e.,
    #             #
    #             #    I ~ |xj-x0| * (xj-x0) . A( 0.5*(xj+x0) ).
    #             #
    #             node0 = self.mesh.nodes[edge[0]]
    #             node1 = self.mesh.nodes[edge[1]]
    #             midpoint = 0.5 * (node0 + node1)

    #             # Instead of projecting onto the normalized edge and then
    #             # multiplying with the edge length for the approximation of
    #             # the integral, just project on the not normalized edge.
    #             a_integral = numpy.dot(
    #                 node1 - node0,
    #                 self._magnetic_vector_potential(midpoint)
    #                 )

    #             # sum it in
    #             sum += 2.0 * self._coedge_edge_ratios[k][l] * \
    #                 (1.0 - math.cos(a_integral))
    #             l += 1
    #         k += 1

    #     return sum / len(self.mesh.nodes)
